// © 2021 Dag Langmyhr, Institutt for informatikk, Universitetet i Oslo

package no.uio.ifi.asp.runtime;

import java.util.ArrayList;
import java.util.Scanner;
import java.util.NoSuchElementException;

import no.uio.ifi.asp.main.*;
import no.uio.ifi.asp.parser.AspSyntax;

public class RuntimeLibrary extends RuntimeScope {
    private Scanner keyboard = new Scanner(System.in);

    public RuntimeLibrary() {
	// float
	assign("float", new RuntimeFunc("float") {
		@Override
		public RuntimeValue evalFuncCall(ArrayList<RuntimeValue> actualParams,
		        AspSyntax where) {
		    checkNumParams(actualParams, 1, "float", where);
		    RuntimeValue arg = actualParams.get(0);
		    if (arg instanceof RuntimeIntValue) {
			long v = arg.getIntValue("parameter to float",where);
			return new RuntimeFloatValue((double)v);
		    } else if (arg instanceof RuntimeFloatValue) {
			return arg;
		    } else if (arg instanceof RuntimeStringValue) {
			String s = arg.getStringValue("parameter to float",where);
			try {
			    double v = Double.parseDouble(s);
			    return new RuntimeFloatValue(v);
			} catch (NumberFormatException e) {
			    runtimeError("String '"+s+"' is not a legal float", where);
			}
		    } 
		    runtimeError("Type error: parameter to float is " +
				 "neither number nor text string", where);
		    return null;
		}});

	// input
	assign("input", new RuntimeFunc("input") {
		@Override
		public RuntimeValue evalFuncCall(ArrayList<RuntimeValue> actualParams,
		        AspSyntax where) {
		    checkNumParams(actualParams, 1, "input", where);
		    String prompt = 
			actualParams.get(0).getStringValue("parameter to input",where);
		    System.out.print(prompt);
		    String line = "";
		    try {
			line = keyboard.nextLine();
		    } catch (IllegalStateException e) {
			Main.panic("Input stream has been closed!", where.lineNum);
		    } catch (NoSuchElementException e) {
			runtimeError("No more lines to read", where);
		    }
		    return new RuntimeStringValue(line);
		}});

	// int
	assign("int", new RuntimeFunc("int") {
		@Override
		public RuntimeValue evalFuncCall(ArrayList<RuntimeValue> actualParams,
		        AspSyntax where) {
		    checkNumParams(actualParams, 1, "int", where);
		    RuntimeValue arg = actualParams.get(0);
		    if (arg instanceof RuntimeIntValue) {
			return arg;
		    } else if (arg instanceof RuntimeFloatValue) {
			double v = arg.getFloatValue("parameter to int",where);
			return new RuntimeIntValue((long)v);
		    } else if (arg instanceof RuntimeStringValue) {
			String s = arg.getStringValue("parameter to int",where);
			try {
			    long v = Long.parseLong(s);
			    return new RuntimeIntValue(v);
			} catch (NumberFormatException e) {
			    runtimeError("String '"+s+"' is not a legal int", where);
			}
		    } 
		    runtimeError("Type error: parameter to int is " +
				 "neither number nor text string", where);
		    return null;
		}});

	// len
	assign("len", new RuntimeFunc("len") {
		@Override
		public RuntimeValue evalFuncCall(
			ArrayList<RuntimeValue> actualParams,
                        AspSyntax where) {
		    checkNumParams(actualParams, 1, "len", where);
		    return actualParams.get(0).evalLen(where);
		}});

	// print
	assign("print", new RuntimeFunc("print") {
		@Override
		public RuntimeValue evalFuncCall(ArrayList<RuntimeValue> actualParams,
		        AspSyntax where) {
		    for (int i = 0;  i < actualParams.size();  ++i) {
			if (i > 0) System.out.print(" ");
			System.out.print(actualParams.get(i).toString());
		    }
		    System.out.println();
		    return new RuntimeNoneValue();
		}});

	// range
	assign("range", new RuntimeFunc("range") {
		@Override
		public RuntimeValue evalFuncCall(ArrayList<RuntimeValue> actualParams,
			AspSyntax where) {
		    checkNumParams(actualParams, 2, "range", where);
		    long v1 = actualParams.get(0).
			getIntValue("1st parameter to range",where);
		    long v2 = actualParams.get(1).
			getIntValue("2nd parameter to range",where);
		    RuntimeListValue res = new RuntimeListValue();
		    for (long i = v1;  i < v2;  ++i)
			res.addElem(new RuntimeIntValue(i));
		    return res;
		}});

	// str
	assign("str", new RuntimeFunc("str") {
		@Override
		public RuntimeValue evalFuncCall(ArrayList<RuntimeValue> actualParams,
		        AspSyntax where) {
		    checkNumParams(actualParams, 1, "str", where);
		    return new RuntimeStringValue(actualParams.get(0).toString());
		}});

    }


    private void checkNumParams(ArrayList<RuntimeValue> actArgs, 
				int nCorrect, String id, AspSyntax where) {
	if (actArgs.size() != nCorrect)
	    RuntimeValue.runtimeError("Wrong number of parameters to "+id+"!",where);
    }
}
