// © 2021 Dag Langmyhr, Institutt for informatikk, Universitetet i Oslo

package no.uio.ifi.asp.scanner;

import java.io.*;
import java.util.*;

import no.uio.ifi.asp.main.*;
import static no.uio.ifi.asp.scanner.TokenKind.*;

public class Scanner {
    private LineNumberReader sourceFile = null;
    private String curFileName;
    private ArrayList<Token> curLineTokens = new ArrayList<>();
    private Stack<Integer> indents = new Stack<>();
    private final int TABDIST = 4;


    public Scanner(String fileName) {
	curFileName = fileName;
	indents.push(0);

	try {
	    sourceFile = new LineNumberReader(
			    new InputStreamReader(
				new FileInputStream(fileName),
				"UTF-8"));
	} catch (IOException e) {
	    scannerError("Cannot read " + fileName + "!");
	}
    }


    private void scannerError(String message) {
	String m = "Asp scanner error";
	if (curLineNum() > 0)
	    m += " on line " + curLineNum();
	m += ": " + message;

	Main.error(m);
    }


    public Token curToken() {
	while (curLineTokens.isEmpty()) {
	    readNextLine();
	}
	return curLineTokens.get(0);
    }


    public void readNextToken() {
	if (! curLineTokens.isEmpty())
	    curLineTokens.remove(0);
    }


    private void readNextLine() {
	curLineTokens.clear();

	// Read the next line:
	String line = null;
	try {
	    line = sourceFile.readLine();
	    if (line == null) {
		sourceFile.close();
		sourceFile = null;
	    } else {
		Main.log.noteSourceLine(curLineNum(), line);
	    }
	} catch (IOException e) {
	    sourceFile = null;
	    scannerError("Unspecified I/O error!");
	}

	// Were there any more lines to read?
	if (sourceFile == null) {
	    while (indents.size() > 1) {
		Token t = new Token(dedentToken);
		curLineTokens.add(t);  Main.log.noteToken(t);
		indents.pop();
	    }

	    Token t = new Token(eofToken);
	    curLineTokens.add(t);  Main.log.noteToken(t);
	    return;
	}

	// Add a few spaces to make the decoding more robust:
	line += "   ";

	// Is it a blank line?
	boolean allBlank = true;
	for (char c: line.toCharArray()) {
	    if (c == '#') break;
	    if (! Character.isWhitespace(c)) {
		allBlank = false;  break;
	    }
	}
	if (allBlank) return;

	// Determine indentation:
	line = expandLeadingTabs(line);
	int ind = findIndent(line);
	if (ind > indents.peek()) {
	    indents.push(ind);
	    curLineTokens.add(new Token(indentToken,curLineNum()));
	} else {
	    while (ind < indents.peek()) {
		curLineTokens.add(new Token(dedentToken, curLineNum()));
		indents.pop();
	    }
	    if (ind != indents.peek()) 
		scannerError("Indentation error!");
	}

	// Find all the tokens:
	int pos = ind;
	while (pos < line.length()) {
	    char c = line.charAt(pos++);

	    if (Character.isWhitespace(c)) {
		// Ignore spaces.
	    } else if (c == '#') {
		// The rest of the line is a comment.
		pos = line.length();
	    } else if (isLetterAZ(c)) {
		String id = "";
		while (true) {
		    id += c;
		    c = line.charAt(pos);
		    if (!isLetterAZ(c) && !isDigit(c)) break;
		    pos++;
		}
		Token t = new Token(nameToken,curLineNum());
		t.name = id;  t.checkResWords();
		curLineTokens.add(t);
	    } else if (isDigit(c)) {
		String n = "" + c;
		while (isDigit(line.charAt(pos))) {
		    n += line.charAt(pos++);
		}

		while (n.length()>1 && n.charAt(0)=='0') {
		    Token t = new Token(integerToken,curLineNum());
		    t.integerLit = 0;
		    curLineTokens.add(t);
		    n = n.substring(1);
		}
		
		if (line.charAt(pos)=='.' && isDigit(line.charAt(pos+1))) {
		    n += '.';  pos++;
		    while (isDigit(line.charAt(pos))) {
			n += line.charAt(pos++);
		    }
		    Token t = new Token(floatToken,curLineNum());
		    t.floatLit = Double.parseDouble(n);
		    curLineTokens.add(t);
		} else {
		    Token t = new Token(integerToken,curLineNum());
		    t.integerLit = Long.parseLong(n);
		    curLineTokens.add(t);
		}
	    } else if (c=='"') {
		String s = "";
		while (true) {
		    if (pos >= line.length())
			scannerError("String literal not terminated!");
		    if (line.charAt(pos) == '"') break;
		    s += line.charAt(pos++);
		}
		Token t = new Token(stringToken,curLineNum());
		t.stringLit = s;
		curLineTokens.add(t);
		pos++;
	    } else if (c=='\'') {
		String s = "";
		while (true) {
		    if (pos >= line.length())
			scannerError("String literal not terminated!");
		    if (line.charAt(pos) == '\'') break;
		    s += line.charAt(pos++);
		}
		Token t = new Token(stringToken,curLineNum());
		t.stringLit = s;
		curLineTokens.add(t);
		pos++;
	    } else if (c=='!' && line.charAt(pos)=='=') {
		curLineTokens.add(new Token(notEqualToken,curLineNum()));
		pos++;
	    } else if (c == '%') {
		curLineTokens.add(new Token(percentToken,curLineNum()));
	    } else if (c == '(') {
		curLineTokens.add(new Token(leftParToken,curLineNum()));
	    } else if (c == ')') {
		curLineTokens.add(new Token(rightParToken,curLineNum()));
	    } else if (c == '*') {
		curLineTokens.add(new Token(astToken,curLineNum()));
	    } else if (c == '+') {
		curLineTokens.add(new Token(plusToken,curLineNum()));
	    } else if (c == ',') {
		curLineTokens.add(new Token(commaToken,curLineNum()));
	    } else if (c == '-') {
		curLineTokens.add(new Token(minusToken,curLineNum()));
	    } else if (c=='/' && line.charAt(pos)=='/') {
		curLineTokens.add(new Token(doubleSlashToken,curLineNum()));
		pos++;
	    } else if (c == '/') {
		curLineTokens.add(new Token(slashToken,curLineNum()));
	    } else if (c == ':') {
		curLineTokens.add(new Token(colonToken,curLineNum()));
	    } else if (c == ';') {
		curLineTokens.add(new Token(semicolonToken,curLineNum()));
	    } else if (c=='<' && line.charAt(pos)=='=') {
		curLineTokens.add(new Token(lessEqualToken,curLineNum()));
		pos++;
	    } else if (c == '<') {
		curLineTokens.add(new Token(lessToken,curLineNum()));
	    } else if (c=='=' && line.charAt(pos)=='=') {
		curLineTokens.add(new Token(doubleEqualToken,curLineNum()));
		pos++;
	    } else if (c == '=') {
		curLineTokens.add(new Token(equalToken,curLineNum()));
	    } else if (c=='>' && line.charAt(pos)=='=') {
		curLineTokens.add(new Token(greaterEqualToken,curLineNum()));
		pos++;
	    } else if (c == '>') {
		curLineTokens.add(new Token(greaterToken,curLineNum()));
	    } else if (c == '[') {
		curLineTokens.add(new Token(leftBracketToken,curLineNum()));
	    } else if (c == ']') {
		curLineTokens.add(new Token(rightBracketToken,curLineNum()));
	    } else if (c == '{') {
		curLineTokens.add(new Token(leftBraceToken,curLineNum()));
	    } else if (c == '}') {
		curLineTokens.add(new Token(rightBraceToken,curLineNum()));
	    } else {
		scannerError("Illegal character: '" + c + "'!");
	    }
	}

	// Terminate line:
	curLineTokens.add(new Token(newLineToken,curLineNum()));

	for (Token t: curLineTokens) 
	    Main.log.noteToken(t);
    }

    public int curLineNum() {
	return sourceFile!=null ? sourceFile.getLineNumber() : 0;
    }

    private int findIndent(String s) {
	int indent = 0;

	while (indent<s.length() && s.charAt(indent)==' ') indent++;
	return indent;
    }

    private String expandLeadingTabs(String s) {
	String newS = "";
	for (int i = 0;  i < s.length();  i++) {
	    char c = s.charAt(i);
	    if (c == '\t') {
		do {
		    newS += " ";
		} while (newS.length()%TABDIST > 0);
	    } else if (c == ' ') {
		newS += " ";
	    } else {
		newS += s.substring(i);
		break;
	    }
	}
	return newS;
    }


    private boolean isLetterAZ(char c) {
	return ('A'<=c && c<='Z') || ('a'<=c && c<='z') || (c=='_');
    }


    private boolean isDigit(char c) {
	return '0'<=c && c<='9';
    }


    public boolean isCompOpr() {
	TokenKind k = curToken().kind;
	return k==doubleEqualToken || k==greaterToken || 
	    k==greaterEqualToken || k==lessToken || 
	    k==lessEqualToken || k==notEqualToken;
    }


    public boolean isFactorPrefix() {
	TokenKind k = curToken().kind;
	return k==minusToken || k==plusToken;
    }


    public boolean isFactorOpr() {
	TokenKind k = curToken().kind;
	return k==astToken || k==doubleSlashToken || 
	    k==percentToken || k==slashToken;
    }
	

    public boolean isTermOpr() {
	TokenKind k = curToken().kind;
	return k==minusToken || k==plusToken;
    }


    public boolean anyEqualToken() {
	for (Token t: curLineTokens) {
	    if (t.kind == equalToken) return true;
	    if (t.kind == semicolonToken) return false;
	}
	return false;
    }
}
